package com.vazil.vridge.docs.service;

import com.vazil.vridge.docs.model.GuideIndex;
import com.vazil.vridge.docs.repository.GuideIndexRepository;
import com.vazil.vridge.docs.utils.TimeManager;
import lombok.AllArgsConstructor;
import lombok.extern.log4j.Log4j2;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.List;
import java.util.NoSuchElementException;

@Service
@AllArgsConstructor
@Log4j2
public class GuideIndexService {
    GuideIndexRepository repository;

    @Transactional
    public GuideIndex getById(String id) throws Exception{
        GuideIndex db = repository.findById(id).orElseThrow(NoSuchElementException::new);

        int viewCount = db.getView() != null ? db.getView() : 0;
        db.setView(viewCount + 1);
        repository.save(db);

        db.setNextIndex(getNextIndex(db, false));
        db.setPrevIndex(getPrevIndex(db));

        return db;
    }

    public GuideIndex getPrevIndex(GuideIndex targetIndex) throws Exception{
        String path = targetIndex.getPath();
        String parentPath = path.substring(0, path.lastIndexOf("/", path.length() -2) + 1);

        // depth == 1, order == 0 인 경우 이전 인덱스 없음
        if(targetIndex.getDepth() == 1 && targetIndex.getOrder() == 0)
            return null;

        // order == 0 인 경우 부모 인덱스 반환
        if(targetIndex.getOrder() == 0){
            return repository.findByPath(parentPath);
        }

        // 이전 형제의 자식이 없을 경우 해당 형제 반환
        GuideIndex prevSibling = repository.findByPathContainingAndDepthAndOrder(parentPath, targetIndex.getDepth(), targetIndex.getOrder() - 1);
        if(repository.findAllByPathContainingAndDepth(prevSibling.getPath(), prevSibling.getDepth() + 1).size() == 0){
            return prevSibling;
        } else { // 이전 형제의 자식이 있을 경우 가장 마지막 자식 반환
            return getLastChildIndex(prevSibling);
        }
    }

    public GuideIndex getLastChildIndex(GuideIndex targetIndex) throws Exception{
        GuideIndex lastChild = repository.findAllByPathContainingAndDepthOrderByOrderDesc(targetIndex.getPath(), targetIndex.getDepth() + 1).get(0);

        if(repository.findAllByPathContainingAndDepth(lastChild.getPath(), lastChild.getDepth() + 1).size() == 0){
            return lastChild;
        } else {
            return getLastChildIndex(lastChild);
        }
    }


    /**
     * @param targetIndex
     * @param onlySibling true일 경우, 자식 guide는 무시하고 다음 형제 guide를 찾음
     */
    public GuideIndex getNextIndex(GuideIndex targetIndex, Boolean onlySibling) throws Exception{
        String path = targetIndex.getPath();
        String parentPath = targetIndex.getDepth() == 1 ?
                "/" : path.substring(0, path.lastIndexOf("/", path.length() -2) + 1);
        GuideIndex nextIndex = null;

        // child가 있으면 order == 0 인 자식 반환 (onlySibling == true 일 경우 패스하여 형제 반환으로 넘어감)
        if(!onlySibling){
            nextIndex = repository.findByPathContainingAndDepthAndOrder(path, targetIndex.getDepth() + 1, 0);
        }

        // child가 없으면 다음 형제 반환
        if(nextIndex == null){
            nextIndex = repository.findByPathContainingAndDepthAndOrder(parentPath, targetIndex.getDepth(), targetIndex.getOrder() + 1);
        }

        // child가 없고, 다음 형제 없는 경우, 부모 guide의 nextIndex 탐색
        if(nextIndex == null && targetIndex.getDepth() != 1) {
            GuideIndex parentIndex = repository.findByPath(parentPath);
            nextIndex = getNextIndex(parentIndex, true);
        }

        return nextIndex;
    }

    public List<GuideIndex> getAll() throws Exception{
        return repository.findAllByOrderByOrder();
    }

    @Transactional
    public GuideIndex create(GuideIndex index) throws Exception{
        autoUpdateOrder(index, 1);

        index.setLike(0);
        index.setView(0);
        index.setCreateDate(TimeManager.now());
        return repository.save(index);
    }

    @Transactional
    public GuideIndex update(GuideIndex index) throws Exception{
        GuideIndex db = repository.findById(index.getId()).orElseThrow(NoSuchElementException::new);

        autoUpdateOrder(db, -1);
        autoUpdateOrder(index, 1);

        db.setLocale(index.getLocale());
        db.setTitle(index.getTitle());
        db.setOrder(index.getOrder());
        db.setPath(index.getPath());
        db.setDepth(index.getDepth());

        return repository.save(db);
    }

    @Transactional
    public String delete(String id) throws Exception{
        GuideIndex db = repository.findById(id).orElseThrow(NoSuchElementException::new);
        autoUpdateOrder(db, -1);
        repository.delete(db);
        return db.getId();
    }


    // targetIndex의 형제들 order 업데이트
    @Transactional
    public void autoUpdateOrder(GuideIndex targetIndex, Integer type) throws Exception {
        String path = targetIndex.getPath();
        String parentPath = path.substring(0, path.lastIndexOf("/", path.length() -2) + 1);

        // 인덱스 증가해야 할 경우 (type == 1)
        if(type == 1){
            if(targetIndex.getDepth() == 1){
                List<GuideIndex> siblings = repository.findAllByDepthOrderByOrder(1);
                siblings.forEach(el -> {
                    if(el.getOrder() >= targetIndex.getOrder()){
                        el.setOrder(el.getOrder() + 1);
                        repository.save(el);
                    }
                });
            } else {
                List<GuideIndex> siblings = repository.findAllByPathContainingAndDepth(parentPath, targetIndex.getDepth());
                siblings.forEach(el -> {
                    if(el.getOrder() >= targetIndex.getOrder()){
                        el.setOrder(el.getOrder() + 1);
                        repository.save(el);
                    }
                });
            }
            return ;
        }


        // 인덱스 감소해야 할 경우 (type == -1)
        if(type == -1){
            if(targetIndex.getDepth() == 1){
                List<GuideIndex> siblings = repository.findAllByDepthOrderByOrder(1);
                siblings.forEach(el -> {
                    if(el.getOrder() > targetIndex.getOrder()){
                        el.setOrder(el.getOrder() - 1);
                        repository.save(el);
                    }
                });
            } else {
                List<GuideIndex> siblings = repository.findAllByPathContainingAndDepth(parentPath, targetIndex.getDepth());
                siblings.forEach(el -> {
                    if(el.getOrder() > targetIndex.getOrder()){
                        el.setOrder(el.getOrder() - 1);
                        repository.save(el);
                    }
                });
            }
            return ;
        }

        log.info("type 파라미터가 부적절합니다.");
    }
}
